@@ -6,7 +6,7 @@ use Carp;
 use integer;
 use vars qw( @s1 @s2 @s3 @s4 @s5 @s6 @s7 @s8 $VERSION );
 
-$VERSION = 0.02;
+$VERSION = 0.03;
 
 sub new {
   my ($class, $key) = @_;
@@ -114,7 +114,7 @@ modify it under the same terms as Perl itself.
 
 =cut
 
-1 # end CAST5_PP.pm
+1 # end module
 __END__
 
 # Delay loading s-boxes until first time init() is called.
@@ -487,7 +487,6 @@ sub init {
   my ($cast5, $key) = @_;
   croak "Key length must be 40 to 128 bits"
       if length($key) < 5 || length($key) > 16;
-  $cast5->{rounds} = length($key) <= 10 ? 12 : 16;
 
   # null-pad the key to 16 bytes, and split it into 32-bit chunks
   my ($s, $t, $u, $v) = unpack "N4", pack "a16", $key;
@@ -531,6 +530,9 @@ sub init {
   }
 
   for (16..31) { $k[$_] &= 31 }
+  delete $cast5->{encrypt};
+  delete $cast5->{decrypt};
+  $cast5->{rounds} = length($key) <= 10 ? 12 : 16;
   $cast5->{key} = \@k;
   return $cast5;
 } # init
@@ -539,64 +541,72 @@ sub encrypt {
   use integer;
   my ($cast5, $block) = @_;
   croak "Block size must be 8" if length($block) != 8;
-  my ($l, $r) = unpack "N2", $block;
-  my ($i, $j);
-  my $k = $cast5->{key} or croak "Call init() first";
 
-  # encrypt the data
-  # don't try to read this -- it's generated by mkencrypt
-  $i=$k->[0]+$r;$j=$k->[16];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]^$s2[$i>>16&255])-$s3[$i>>8&255])+$s4[$i&255];
-  $i=$k->[1]^$l;$j=$k->[17];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]-$s2[$i>>16&255])+$s3[$i>>8&255])^$s4[$i&255];
-  $i=$k->[2]-$r;$j=$k->[18];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]+$s2[$i>>16&255])^$s3[$i>>8&255])-$s4[$i&255];
-  $i=$k->[3]+$l;$j=$k->[19];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]^$s2[$i>>16&255])-$s3[$i>>8&255])+$s4[$i&255];
-  $i=$k->[4]^$r;$j=$k->[20];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]-$s2[$i>>16&255])+$s3[$i>>8&255])^$s4[$i&255];
-  $i=$k->[5]-$l;$j=$k->[21];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]+$s2[$i>>16&255])^$s3[$i>>8&255])-$s4[$i&255];
-  $i=$k->[6]+$r;$j=$k->[22];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]^$s2[$i>>16&255])-$s3[$i>>8&255])+$s4[$i&255];
-  $i=$k->[7]^$l;$j=$k->[23];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]-$s2[$i>>16&255])+$s3[$i>>8&255])^$s4[$i&255];
-  $i=$k->[8]-$r;$j=$k->[24];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]+$s2[$i>>16&255])^$s3[$i>>8&255])-$s4[$i&255];
-  $i=$k->[9]+$l;$j=$k->[25];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]^$s2[$i>>16&255])-$s3[$i>>8&255])+$s4[$i&255];
-  $i=$k->[10]^$r;$j=$k->[26];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]-$s2[$i>>16&255])+$s3[$i>>8&255])^$s4[$i&255];
-  $i=$k->[11]-$l;$j=$k->[27];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]+$s2[$i>>16&255])^$s3[$i>>8&255])-$s4[$i&255];
-  if ($cast5->{rounds} == 16) {
-    $i=$k->[12]+$r;$j=$k->[28];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]^$s2[$i>>16&255])-$s3[$i>>8&255])+$s4[$i&255];
-    $i=$k->[13]^$l;$j=$k->[29];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]-$s2[$i>>16&255])+$s3[$i>>8&255])^$s4[$i&255];
-    $i=$k->[14]-$r;$j=$k->[30];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]+$s2[$i>>16&255])^$s3[$i>>8&255])-$s4[$i&255];
-    $i=$k->[15]+$l;$j=$k->[31];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]^$s2[$i>>16&255])-$s3[$i>>8&255])+$s4[$i&255];
+  my $encrypt = $cast5->{encrypt};
+  unless ($encrypt) {
+    my $key = $cast5->{key} or croak "Call init() first";
+    $f = 'sub{my($l,$r,$i)=unpack"N2",$_[0];';
+
+    my ($l, $r) = qw( $l $r );
+    my ($op1, $op2, $op3) = qw( + ^ - );
+    foreach my $round (0 .. $cast5->{rounds}-1) {
+      my $km = $key->[$round];
+      my $kr = $key->[$round+16];
+
+      my $rot = "";
+      if ($kr) {
+        my $mask = ~(~0<<$kr) & 0xffffffff;
+        my $kr2 = 32-$kr;
+        $rot = "\$i=\$i<<$kr|\$i>>$kr2&$mask;"
+      }
+
+      $f .= "\$i=$km$op1$r;$rot$l^=((\$s1[\$i>>24&255]$op2\$s2[\$i>>16&255])$op3\$s3[\$i>>8&255])$op1\$s4[\$i&255];";
+      ($l, $r) = ($r, $l);
+      ($op1, $op2, $op3) = ($op2, $op3, $op1);
+    }
+
+    $f .= 'pack"N2",$r&0xffffffff,$l&0xffffffff}';
+    $cast5->{encrypt} = $encrypt = eval $f;
   }
 
-  return pack "N2", $r, $l;
+  return &$encrypt($block);
 } # encrypt
 
 sub decrypt {
   use integer;
   my ($cast5, $block) = @_;
   croak "Block size must be 8" if length($block) != 8;
-  my ($r, $l) = unpack "N2", $block;
-  my ($i, $j);
-  my $k = $cast5->{key} or croak "Call init() first";
 
-  # decrypt the data
-  # don't try to read this -- it's generated by mkdecrypt
-  if ($cast5->{rounds} == 16) {
-    $i=$k->[15]+$l;$j=$k->[31];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]^$s2[$i>>16&255])-$s3[$i>>8&255])+$s4[$i&255];
-    $i=$k->[14]-$r;$j=$k->[30];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]+$s2[$i>>16&255])^$s3[$i>>8&255])-$s4[$i&255];
-    $i=$k->[13]^$l;$j=$k->[29];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]-$s2[$i>>16&255])+$s3[$i>>8&255])^$s4[$i&255];
-    $i=$k->[12]+$r;$j=$k->[28];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]^$s2[$i>>16&255])-$s3[$i>>8&255])+$s4[$i&255];
+  my $decrypt = $cast5->{decrypt};
+  unless ($decrypt) {
+    my $key = $cast5->{key} or croak "Call init() first";
+    my $rounds = $cast5->{rounds};
+    $f = 'sub{my($r,$l,$i)=unpack"N2",$_[0];';
+
+    my ($l, $r) = qw( $r $l );
+    my ($op1, $op2, $op3) = qw( - + ^ );
+    foreach (1 .. $rounds%3) { ($op1, $op2, $op3) = ($op2, $op3, $op1) }
+    foreach my $round (1 .. $rounds) {
+      my $km = $key->[$rounds-$round];
+      my $kr = $key->[$rounds-$round+16];
+
+      my $rot = "";
+      if ($kr) {
+        my $mask = ~(~0<<$kr) & 0xffffffff;
+        my $kr2 = 32-$kr;
+        $rot = "\$i=\$i<<$kr|\$i>>$kr2&$mask;"
+      }
+
+      $f .= "\$i=$km$op1$r;$rot$l^=((\$s1[\$i>>24&255]$op2\$s2[\$i>>16&255])$op3\$s3[\$i>>8&255])$op1\$s4[\$i&255];";
+      ($l, $r) = ($r, $l);
+      ($op1, $op2, $op3) = ($op3, $op1, $op2);
+    }
+
+    $f .= 'pack"N2",$l&0xffffffff,$r&0xffffffff}';
+    $cast5->{decrypt} = $decrypt = eval $f;
   }
-  $i=$k->[11]-$l;$j=$k->[27];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]+$s2[$i>>16&255])^$s3[$i>>8&255])-$s4[$i&255];
-  $i=$k->[10]^$r;$j=$k->[26];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]-$s2[$i>>16&255])+$s3[$i>>8&255])^$s4[$i&255];
-  $i=$k->[9]+$l;$j=$k->[25];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]^$s2[$i>>16&255])-$s3[$i>>8&255])+$s4[$i&255];
-  $i=$k->[8]-$r;$j=$k->[24];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]+$s2[$i>>16&255])^$s3[$i>>8&255])-$s4[$i&255];
-  $i=$k->[7]^$l;$j=$k->[23];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]-$s2[$i>>16&255])+$s3[$i>>8&255])^$s4[$i&255];
-  $i=$k->[6]+$r;$j=$k->[22];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]^$s2[$i>>16&255])-$s3[$i>>8&255])+$s4[$i&255];
-  $i=$k->[5]-$l;$j=$k->[21];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]+$s2[$i>>16&255])^$s3[$i>>8&255])-$s4[$i&255];
-  $i=$k->[4]^$r;$j=$k->[20];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]-$s2[$i>>16&255])+$s3[$i>>8&255])^$s4[$i&255];
-  $i=$k->[3]+$l;$j=$k->[19];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]^$s2[$i>>16&255])-$s3[$i>>8&255])+$s4[$i&255];
-  $i=$k->[2]-$r;$j=$k->[18];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]+$s2[$i>>16&255])^$s3[$i>>8&255])-$s4[$i&255];
-  $i=$k->[1]^$l;$j=$k->[17];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$r^=(($s1[$i>>24&255]-$s2[$i>>16&255])+$s3[$i>>8&255])^$s4[$i&255];
-  $i=$k->[0]+$r;$j=$k->[16];{no integer;$i=($i<<$j)|(($i&0xffffffff)>>(32-$j))}$l^=(($s1[$i>>24&255]^$s2[$i>>16&255])-$s3[$i>>8&255])+$s4[$i&255];
 
-  return pack "N2", $l, $r;
+  return &$decrypt($block);
 } # decrypt
 
 # end CAST5_PP.pm
